# SPDX-License-Identifier: LGPL-3.0-only AND Apache-2.0
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>
# Copyright 2015 The Prometheus Authors
# Portions of this file are adapted from prometheus_client

import bisect
from contextlib import contextmanager
import itertools
import multiprocessing
import threading
import time

from prometheus_client import metrics_core, registry

import _mpmetrics
from .atomic import AtomicUInt64, AtomicDouble
from .generics import IntType
from .heap import Heap
from .types import Box, Double, Array, Struct
from .util import classproperty, genmask

@contextmanager
def Timer(callback):
    now = time.perf_counter()
    yield
    callback(max(time.perf_counter() - now, 0))

class Collector:
    def __init__(self, metric, name, docs, registry, heap, kwargs):
        self._name = name
        self._docs = docs
        self._metric = metric(heap, **kwargs)
        
        registry.register(self)

    def __getattr__(self, name):
        return getattr(self.__dict__['_metric'], name)

    def _family(self):
        return metrics_core.Metric(self._name, self._docs, self._metric.typ)

    def describe(self):
        yield self._family()

    def collect(self):
        family = self._family()
        def add_sample(suffix, value, labels={}):
            family.add_sample(self._name + suffix, labels, value)
        self._metric._sample(add_sample)
        yield family

class LabeledCollector:
    _manager_lock = threading.Lock()

    @classproperty
    def manager(cls):
        with cls._manager_lock:
            if not hasattr(cls, '_manager'):
                cls._manager = multiprocessing.Manager()
            return cls._manager

    def __init__(self, metric, name, docs, labelnames, registry, heap, kwargs):
        self._metric = metric
        self._name = name
        self._docs = docs
        self._kwargs = kwargs
        self._heap = heap

        self._labelnames = tuple(labelnames)
        for label in self._labelnames:
            if not metrics_core.METRIC_LABEL_NAME_RE.match(label):
                raise ValueError(f"invalid label {label}")

            if metrics_core.RESERVED_METRIC_LABEL_NAME_RE.match(label):
                raise ValueError(f"reserved label {label}")

            if hasattr(self._metric, 'reserved_labels') and label in self._metric.reserved_labels:
                raise ValueError(f"reserved label {label}")

        self._lock = threading.Lock()
        self._cache = dict()

        self._shared_lock = Box[_mpmetrics.Lock](self._heap)
        self._metrics = getattr(registry, 'manager', self.manager).dict()
        registry.register(self)

    def _label_values(self, values, labels):
        if values and labels:
            raise ValueError("can't pass both *args and **kwargs")

        if labels:
            if sorted(labels) != sorted(self._labelnames):
                raise ValueError("incorrect label names")
            return tuple(str(labels[label]) for label in self._labelnames)
        else:
            if len(values) != len(self._labelnames):
                raise ValueError("incorrect label count")
            return tuple(str(label) for label in values)

    def labels(self, *values, **labels):
        values = self._label_values(values, labels)

        with self._lock:
            metric = self._cache.get(values)
            if not metric:
                with self._shared_lock:
                    metric = self._metrics.get(values)
                    if not metric:
                        metric = self._metric(self._heap, **self._kwargs)
                        self._metrics[values] = metric
                self._cache[values] = metric
            return metric

    def _family(self):
        return metrics_core.Metric(self._name, self._docs, self._metric.typ)

    def describe(self):
        yield self._family()

    def collect(self):
        family = self._family()
        with self._lock:
            metrics = self._cache = dict(self._metrics)

        for labelvalues, metric in metrics.items():
            metric_labels = dict(zip(self._labelnames, labelvalues))
            def add_sample(suffix, value, labels={}):
                family.add_sample(self._name + suffix, metric_labels | labels, value)
            metric._sample(add_sample)
        yield family

class CollectorFactory:
    _heap_lock = threading.Lock()

    @classproperty
    def heap(cls):
        with cls._heap_lock:
            if not hasattr(cls, '_heap'):
                cls._heap = Heap()
            return cls._heap

    def __init__(self, metric):
        self._metric = metric

    def __getattr__(self, name):
        return getattr(self.__dict__['_metric'], name)

    def __call__(self, name, documentation, labelnames=(), namespace="",
                 subsystem="", unit="", registry=registry.REGISTRY, **kwargs):

        parts = []
        if namespace:
            parts.append(namespace)
        if subsystem:
            parts.append(subsystem)
        if self._metric.typ == 'counter':
            name = name.removesuffix('_total')
        if unit:
            name = name.removesuffix('_' + unit)
        parts.append(name)
        if unit:
            if self._metric.typ in ('info', 'stateset'):
                raise ValueError(f"{self._metric.typ} metrics cannot have a unit")
            parts.append(unit)

        name = '_'.join(parts)
        if not metrics_core.METRIC_NAME_RE.match(name):
            raise ValueError(f"invalid metric name {name}")

        heap = getattr(registry, 'heap', self.heap)

        if labelnames:
            return LabeledCollector(self._metric, name, documentation, labelnames, registry, heap,
                                    kwargs)
        return Collector(self._metric, name, documentation, registry, heap, kwargs)

class Counter(Struct):
    typ = 'counter'
    _fields_ = {
        '_total': AtomicUInt64,
        '_created': Double,
    }

    def __init__(self, mem,  init=True):
        super().__init__(mem, init)
        if init:
            self._created.value = time.time()

    def inc(self, amount=1, exemplar=None):
        if amount < 0:
            raise ValueError("amount must be positive")

        if exemplar is not None:
            raise NotImplementedError("exemplars are not yet supported")
   
        self._total.add(amount)

    def _sample(self, add_sample):
        add_sample('_total', self._total.get())
        add_sample('_created', self._created.value)

    @contextmanager
    def count_exceptions(self, exception=Exception):
        try:
            yield
        except exception:
            self.inc()

Counter = CollectorFactory(Box[Counter])

class Gauge:
    typ = 'gauge'

    def __init__(self, heap):
        self._value = Box[AtomicDouble](heap)

    def inc(self, amount=1):
        self._value.add(amount)

    def dec(self, amount=1):
        self._value.add(-amount)

    def set(self, amount):
        self._value.set(amount)

    def _sample(self, add_sample):
        add_sample('', self._value.get())

    def set_to_current_time(self):
        self.set(time.time())

    @contextmanager
    def track_inprogress(self):
        self.inc()
        try:
            yield
        finally:
            self.dec()

    def time(self):
        return Timer(self.set)

Gauge = CollectorFactory(Gauge)

class _SummaryData(Struct):
    _fields_ = {
        'sum': AtomicDouble,
        'count': AtomicUInt64,
    }

class Summary(Struct):
    typ = 'summary'
    reserved_labels = ('quantile',)
    _fields_ = {
        '_lock': _mpmetrics.Lock,
        '_data': Array[_SummaryData, 2],
        '_count': AtomicUInt64,
        '_created': Double,
    }

    def __init__(self, mem, init=True):
        super().__init__(mem, init)
        if init:
            self._created.value = time.time()

    def observe(self, amount):
        data = self._data[self._count.add(1) >> 63]
        data.sum.add(amount)
        data.count.add(1)

    def _sample(self, add_sample):
        with self._lock:
            count = self._count.add(1 << 63, raise_on_overflow=False)
            hot = self._data[~count >> 63]
            cold = self._data[count >> 63]
            count &= genmask(62, 0)

            while cold.count.get() != count:
                time.sleep(0)

            sum = cold.sum.get()
            hot.count.add(count)
            hot.sum.add(sum)
            cold.count.set(0)
            cold.sum.set(0)

        add_sample('_count', count)
        add_sample('_sum', sum)
        add_sample('_created', self._created.value)

    def time(self):
        return Timer(self.observe)

Summary = CollectorFactory(Box[Summary])

def _HistogramData(__name__, bucket_count):
    _fields_ = {
        'buckets': Array[AtomicUInt64, bucket_count],
        'sum': AtomicDouble,
        'count': AtomicUInt64,
    }

    ns = locals()
    del ns['bucket_count']
    return type(__name__, (Struct,), ns)

_HistogramData = IntType('_HistogramData', _HistogramData)

def _Histogram(__name__, bucket_count):
    typ = 'histogram'
    _fields_ = {
        '_lock': _mpmetrics.Lock,
        '_data': Array[_HistogramData[bucket_count], 2],
        '_count': AtomicUInt64,
        '_created': Double,
    }

    def __init__(self, mem, thresholds, init=True):
        assert len(thresholds) == bucket_count
        self.thresholds = thresholds

        Struct.__init__(self, mem, init)
        if init:
            self._created.value = time.time()

    def observe(self, amount, exemplar=None):
        if exemplar is not None:
            raise NotImplementedError("exemplars are not yet supported")

        i = bisect.bisect_left(self.thresholds, amount)

        data = self._data[self._count.add(1) >> 63]
        data.buckets[i].add(1)
        data.sum.add(amount)
        data.count.add(1)

    def _sample(self, add_sample):
        with self._lock:
            count = self._count.add(1 << 63, raise_on_overflow=False)
            hot = self._data[~count >> 63]
            cold = self._data[count >> 63]
            count &= genmask(62, 0)

            while cold.count.get() != count:
                time.sleep(0)

            buckets = [bucket.get() for bucket in cold.buckets]
            sum = cold.sum.get()

            for i, bucket in enumerate(buckets):
                hot.buckets[i].add(bucket)
            hot.sum.add(sum)
            hot.count.add(count)

            for bucket in cold.buckets:
                bucket.set(0)
            cold.sum.set(0)
            cold.count.set(0)

        for val, le in zip(itertools.accumulate(buckets), self.thresholds):
            add_sample('_bucket', val, { 'le': str(le) })
        add_sample('_sum', sum)
        add_sample('_count', count)
        add_sample('_created', self._created.value)

    ns = locals()
    ns['time'] = lambda self: Timer(self.observe)
    del ns['bucket_count']

    return type(__name__, (Struct,), ns)

_Histogram = IntType('_Histogram', _Histogram)

class _HistogramFactory:
    typ = 'histogram'
    reserved_labels = ('le',)
    DEFAULT_BUCKETS = (.005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0, 7.5, 10.0,
                       float('inf'))

    def __call__(self, heap, buckets=DEFAULT_BUCKETS, **kwargs):
        thresholds = [float(b) for b in buckets]
        if thresholds != sorted(thresholds):
            raise ValueError('thresholds not in sorted order')
        if thresholds and thresholds[-1] != float('inf'):
            thresholds.append(float('inf'))
        if len(thresholds) < 2:
            raise ValueError('must have at least two thresholds')
        thresholds =  tuple(thresholds)

        histogram = Box[_Histogram[len(thresholds)]]
        return histogram(heap, thresholds=thresholds, **kwargs)

Histogram = CollectorFactory(_HistogramFactory())



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import _mpmetrics
from .types import Double, Int64, UInt64, Struct

# TODO: Rewrite this in C if anyone cares about performance on arches without 64-bit atomics?

class _Locking(Struct):
    _fields_ = {
        '_lock': _mpmetrics.Lock,
    }

    def get(self):
        with self._lock:
            return self._value.value

    def set(self, value):
        with self._lock:
            self._value.value = value

    def add(self, amount, raise_on_overflow=True):
        with self._lock:
            old = self._value.value
            self._value.value = old + amount
            if raise_on_overflow and self._value.value != old + amount:
                raise OverflowError(f"{old} + {amount} too large to fit")
            return old

class LockingDouble(_Locking):
    _fields_ = _Locking._fields_ | {
        '_value': Double,
    }

class LockingInt64(_Locking):
    _fields_ = _Locking._fields_ | {
        '_value': Int64,
    }

class LockingUInt64(_Locking):
    _fields_ = _Locking._fields_ | {
        '_value': UInt64,
    }

AtomicDouble = _mpmetrics.AtomicDouble or LockingDouble
AtomicInt64 = _mpmetrics.AtomicInt64 or LockingInt64
AtomicUInt64 = _mpmetrics.AtomicUInt64 or LockingUInt64



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2021-22 Sean Anderson <seanga2@gmail.com>

import itertools
import mmap
import os
from tempfile import NamedTemporaryFile
import threading
from weakref import WeakValueDictionary

import _mpmetrics
from .types import Size_t, Struct
from .util import align, _align_check

SC_LEVEL1_DCACHE_LINESIZE = 190
if (CACHELINESIZE := os.sysconf(SC_LEVEL1_DCACHE_LINESIZE)) < 0:
    CACHELINESIZE = 64 # Assume 64-byte cache lines

PAGESIZE = mmap.ALLOCATIONGRANULARITY

class Heap(Struct):
    _fields_ = {
        '_shared_lock': _mpmetrics.Lock,
        '_base': Size_t,
    }

    # Only create one heap per process to avoid duplicate mappings
    _heaps_lock = threading.Lock()
    _heaps = WeakValueDictionary()

    def __new__(cls, map_size=PAGESIZE, filename=None):
        cls._heaps_lock.acquire()
        if filename:
            heap = cls._heaps.get(filename, None)
            if heap:
                cls._heaps_lock.release()
                return heap

        try:
            return super().__new__(cls)
        except:
            if filename:
                cls._heaps_lock.release()
            raise

    # Must be called with Heap._heaps_lock held; it will be released
    def __init__(self, map_size=PAGESIZE, filename=None):
        try:
            if map_size % PAGESIZE:
                raise ValueError("size must be a multiple of {}".format(PAGESIZE))
            _align_check(map_size)
            self.map_size = map_size

            # File backing our shared memory
            if filename:
                self._file = open(filename, 'a+b') 
            else:
                self._file = NamedTemporaryFile()
            self._fd = self._file.fileno()

            if not filename:
                # Allocate a page to start with
                os.truncate(self._fd, map_size)

            # Process-local shared memory maps
            self._maps = [mmap.mmap(self._fd, map_size)]
            # Lock for _maps
            self._lock = threading.Lock()

            super().__init__(memoryview(self._maps[0])[:self.size], filename)
            if not filename:
                self._base.value = self.size

            # Add ourself to the list of heaps
            self._heaps[self._file.name] = self
        finally:
            self._heaps_lock.release()

    def __getnewargs__(self):
        return self.map_size, self._file.name

    def __getstate__(self):
        return self.map_size, self._file.name

    def __setstate__(self, state):
        map_size, filename = state
        if not hasattr(self, '_file'):
            self.__init__(map_size, filename)

    def __del__(self):
        if hasattr(self, '_file'):
            self._file.close()

    class Block:
        def __init__(self, heap, start, size):
            self.heap = heap
            self.start = start
            self.size = size

        def deref(self):
            heap = self.heap
            page = int(self.start / heap.map_size)
            page_off = page * heap.map_size
            off = self.start - page_off
            with heap._lock:
                if len(heap._maps) <= page:
                    heap._maps.extend(itertools.repeat(None, page - len(heap._maps) + 1))
                if not self.heap._maps[page]:
                    heap._maps[page] = mmap.mmap(heap._fd, heap.map_size, offset=page_off)
                map = heap._maps[page]

            return memoryview(map)[off:off+self.size]

        def free(self):
            pass

    def malloc(self, size, alignment=CACHELINESIZE):
        if size <= 0:
            raise ValueError("size must be strictly positive")
        elif size > self.map_size:
            raise ValueError("size must be less than {}".format(self.map_size))
        _align_check(alignment)

        with self._shared_lock:
            total = align(self._base.value, self.map_size)
            self._base.value = align(self._base.value, alignment)
            if self._base.value + size >= total:
                os.ftruncate(self._fd, total + self.map_size)
                self._base.value = total
            start = self._base.value
            self._base.value += size

        return self.Block(self, start, size)



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

from collections import namedtuple
import multiprocessing
import queue
import threading
import time

import pytest

from mpmetrics.heap import Heap

Parallel = namedtuple('Parallel', ('spawn', 'barrier', 'queue'))
parallels = {
    'process': Parallel(
            spawn = multiprocessing.Process,
            barrier = multiprocessing.Barrier,
            queue = multiprocessing.Queue
        ),
    'thread': Parallel(
            spawn = threading.Thread,
            barrier = threading.Barrier,
            queue = queue.Queue,
        ),
}

@pytest.fixture(scope='session', params=('thread', 'process'))
def parallel(request):
    return parallels[request.param]

class ParallelLoop:
    def __init__(self, parallel, n=4, count=None):
        self.parallel = parallel
        if parallel.spawn is threading.Thread:
            n = n or 2
            count = count or 10000
        else:
            n = n or 4
            count = count or 100000
        self.n = n
        self.count = count
        self.total = n * count
        self.barrier = self.parallel.barrier(self.n + 1)

    def target(self, n):
        if hasattr(self, 'setup'):
            try:
                self.setup(n)
            except:
                self.barrier.abort()
                raise

        self.barrier.wait()
        for i in range(self.count):
            self.loop(i)

    def run(self):
        procs = [self.parallel.spawn(target=self.target, args=(i,)) for i in range(self.n)]
        for p in procs:
            p.start()
        self.barrier.wait()

        try:
            if hasattr(self, 'check'):
                while any(p.is_alive() for p in procs):
                    self.check()
                    time.sleep(0)
        except:
            for p in procs:
                if hasattr(p, 'terminate'):
                    p.terminate()
            raise
        finally:
            for p in procs:
                p.join()

        if hasattr(self, 'final'):
            self.final()

@pytest.fixture(scope="session")
def heap():
    return Heap()



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

def _align_mask(x, mask):
    return (x + mask) & ~mask

def _align_check(a):
    if not a or a & (a - 1):
        raise ValueError("{} is not a power of 2".format(a))

def align(x, a):
    _align_check(a)
    return _align_mask(x, a - 1)

def align_down(x, a):
    _align_check(a)
    return x & ~(a - 1)

def genmask(hi, lo):
    return (-1 << lo) & ~(-1 << (hi + 1))

# From https://stackoverflow.com/a/7864317/5086505
class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import pickle

from hypothesis import assume, given, reject, settings, strategies as st
import pytest

from mpmetrics.atomic import AtomicInt64, AtomicUInt64, AtomicDouble
from mpmetrics.generics import ObjectType, ListType
from mpmetrics.heap import PAGESIZE
from mpmetrics.types import Array, Box, Double, Size_t, Struct
from _mpmetrics import Lock

from .common import heap

def GenericStruct(__name__, classes):
    ns = { '_fields_': { str(i): cls for i, cls in enumerate(classes)}}
    return type(__name__, (Struct,), ns)

GenericStruct = ListType('GenericStruct', GenericStruct, ObjectType) 

def recursive_types(classes):
    @st.composite
    def arrays(draw):
        array = Array[draw(classes), draw(st.integers(1, 8))]
        # reject too-large types as soon as we can
        if array.size >= PAGESIZE:
            reject()
        return array

    @st.composite
    def structs(draw):
        n = draw(st.integers(1, 8))
        struct = GenericStruct[draw(st.tuples(*((classes,) * n)))]
        # ditto
        if struct.size >= PAGESIZE:
            reject()
        return struct

    return arrays() | structs()

base_types = st.sampled_from((AtomicInt64, AtomicUInt64, AtomicDouble, Double, Size_t, Lock))
types = st.recursive(base_types, recursive_types, max_leaves=20)

@given(types)
@settings(max_examples=25)
def test_pickle(heap, cls):
    pickle.loads(pickle.dumps(cls))

    assume(cls.size <= heap.map_size)
    v = Box[cls](heap)
    pickle.loads(pickle.dumps(v))

# We use Size_t because drawing from types is slow
@given(st.integers(min_value=1))
def test_array(heap, n):
    A = Array[Size_t, n]
    assert A.size >= n * Size_t.size

    assume(A.size <= heap.map_size)
    a = Box[A](heap)
    assert len(a) == n

@given(st.integers(max_value=0))
def test_bad_size(n):
    with pytest.raises(ValueError):
        Array[Size_t, n]



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

from .metrics import Counter, Gauge, Summary, Histogram

__all__ = ('Counter', 'Gauge', 'Summary', 'Histogram')



#!/bin/env python3
# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import os
import setuptools

setuptools.setup(
    name = 'mpmetrics',
    use_scm_version = True,
    description = "Multiprocess-safe metrics",
    long_description = open(os.path.join(os.path.dirname(__file__), 'README.md')).read(),
    long_description_content_type = 'text/markdown',
    author = 'Sean Anderson',
    author_email = 'seanga2@gmail.com',
    url = 'https://github.com/Forty-Bot/mpmetrics',
    python_requires = ">=3.9",
    packages = ['mpmetrics'],
    ext_modules = [
        setuptools.Extension(
            '_mpmetrics',
            ['_mpmetrics.c', 'atomic.c', 'lock.c'],
            extra_compile_args = [
                '-Wno-missing-braces',
            ],
        ),
    ],
    license = 'GPL-3.0-only',
    license_files = [
        'COPYING',
        'LICENSES/*',
    ],
    setup_requires = ['setuptools_scm'],
    install_requires = [
        'prometheus_client',
    ],
    extras_require = {
        "tests": [
            'hypothesis',
            'pytest',
        ],
    },
)



# SPDX-License-Identifier: GPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import math
import mmap

import pytest
from hypothesis import given, settings, strategies as st

from mpmetrics.heap import Heap
from mpmetrics.util import align

from .common import parallel

@given(st.integers().filter(lambda n: n % mmap.PAGESIZE))
def test_bad_map_size(map_size):
    with pytest.raises(ValueError):
        Heap(map_size=map_size)

@given(st.integers(max_value=0))
def test_small_size(size):
    with pytest.raises(ValueError):
        Heap().malloc(size)

@given(st.integers(min_value=mmap.PAGESIZE + 1))
def test_big_size(size):
    with pytest.raises(ValueError):
        Heap().malloc(size)

@given(st.integers().filter(lambda a: a <= 0 or math.log2(a) % 1))
def test_bad_align(alignment):
    with pytest.raises(ValueError):
        Heap().malloc(1, alignment)

@st.composite
def allocs(draw):
    size = mmap.ALLOCATIONGRANULARITY << draw(st.integers(0, 8))
    heap = Heap(map_size=size)
    sizes = st.integers(1, size)
    aligns = st.integers(0, 12).map(lambda n: 1 << n)
    return heap, draw(st.lists(st.tuples(sizes, aligns), min_size=3))

@given(allocs())
@settings(max_examples=25)
def test_malloc(alloc):
    h, paramlist = alloc

    blocks = [h.malloc(*params) for params in paramlist]
    for ((size, alignment), block) in zip(paramlist, blocks):
        assert block.size == size
        assert block.start == align(block.start, alignment)

    mems = [block.deref() for block in blocks]
    for (block, mem) in zip(blocks, mems):
        assert len(mem) == block.size
        assert not any(mem)
        mem[:] = b'A' * len(mem)

    blocks.sort(key=lambda block: block.start)
    for i in range(len(blocks)):
        if i:
            prev = blocks[i - 1]
            assert prev.start + prev.size <= blocks[i].start

def test_prefork(parallel):
    mem = Heap().malloc(1).deref()
    assert mem[0] == 0

    def setone(mem):
        mem[0] = 1

    p = parallel.spawn(target=setone, args=(mem,))
    p.start()
    p.join()
    assert mem[0] == 1

def test_postfork(parallel):
    def setone(q):
        mem = q.get().deref()
        mem[0] = 2

    q = parallel.queue()
    p = parallel.spawn(target=setone, args=(q,))
    p.start()

    h = Heap()
    h.malloc(mmap.PAGESIZE)
    block = h.malloc(1)
    q.put(block)
    p.join()
    assert block.deref()[0] == 2



# SPDX-License-Identifier: LGPL-3.0-only
# Copyright (C) 2022 Sean Anderson <seanga2@gmail.com>

import ctypes

from .generics import IntType, ObjectType, ProductType
from .util import align, classproperty

def _wrap_ctype(__name__, ctype):
    size = ctypes.sizeof(ctype)
    align = ctypes.alignment(ctype)

    def __init__(self, mem, init=True):
        self._mem = mem
        self._value = ctype.from_buffer(mem)

    def __getattr__(self, name):
        return getattr(self.__dict__['_value'], name)

    def __setattr__(self, name, value):
        if '_value' in self.__dict__:
            setattr(self.__dict__['_value'], name, value)
        else:
            self.__dict__[name] = value

    def __delattr__(self, name):
        delattr(self.__dict__['_value'], name)

    return type(__name__, (), { name: value for name, value in locals().items()
                                if name != 'ctype'})

Double = _wrap_ctype('Double', ctypes.c_double)
Size_t = _wrap_ctype('Size_t', ctypes.c_size_t)
Int64 = _wrap_ctype('Int64', ctypes.c_int64)
UInt64 = _wrap_ctype('UInt64', ctypes.c_uint64)

class Struct:
    def __init__(self, mem, init=True):
        self._mem = mem
        off = 0
        for name, field in self._fields_.items():
            off = align(off, field.align)
            setattr(self, name, field(mem[off:off + field.size], init))
            off += field.size
            
    @classproperty
    def size(cls):
        size = 0
        for name, field in cls._fields_.items():
            size = align(size, field.align)
            size += field.size
        return size

    @classproperty
    def align(cls):
        return max(field.align for name, field in cls._fields_.items())

def Array(__name__, cls, n):
    if n < 1:
        raise ValueError("n must be strictly positive")

    member_size = align(cls.size, cls.align)
    size = member_size * n

    def __init__(self, mem, init=True):
        self._mem = mem
        self._vals = []
        for i in range(n):
            off = i * member_size
            self._vals.append(cls(mem[off:off + member_size], init))

    def __len__(self):
        return n

    def __getitem__(self, key):
        return self._vals[key]

    def __setitem__(self, key, value):
        self._vals[key] = value

    def __iter__(self):
        return iter(self._vals)

    ns = locals()
    ns['align'] = cls.align
    del ns['member_size']
    del ns['cls']
    del ns['n']

    return type(__name__, (), ns)

Array = ProductType('Array', Array, (ObjectType, IntType))

class _Box:
    def __init__(self, heap, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        block = heap.malloc(self.size)
        super().__init__(block.deref(), init=True, *self._args, **self._kwargs)
        self._block = block

    def __getstate__(self):
        return self._block, self._args, self._kwargs

    def __setstate__(self, state):
        self._block, self._args, self._kwargs = state
        super().__init__(self._block.deref(), init=False, *self._args, **self._kwargs)

Box = ObjectType('Box', lambda name, cls: type(name, (_Box, cls), {}))
